# -*- coding: utf-8 -*-
from __future__ import annotations

import adafruit_as7341
import board
import click
import pioreactor.actions.led_intensity as led_utils
from pioreactor import types as pt
from pioreactor.background_jobs.base import BackgroundJobWithDodgingContrib
from pioreactor.background_jobs.leader.mqtt_to_db_streaming import produce_metadata
from pioreactor.background_jobs.leader.mqtt_to_db_streaming import register_source_to_sink
from pioreactor.background_jobs.leader.mqtt_to_db_streaming import TopicToParserToTable
from pioreactor.config import config
from pioreactor.exc import HardwareNotFoundError
from pioreactor.utils.timing import current_utc_datetime
from pioreactor.whoami import get_latest_experiment_name
from pioreactor.whoami import get_unit_name

# Define the wavelengths and their corresponding LED channels
WAVELENGTH_LED_MAP = {
    415: config.get("spectrometer_reading.config", "led_415", fallback="A"),
    445: config.get("spectrometer_reading.config", "led_445", fallback="A"),
    480: config.get("spectrometer_reading.config", "led_480", fallback="C"),
    515: config.get("spectrometer_reading.config", "led_515", fallback="C"),
    555: config.get("spectrometer_reading.config", "led_555", fallback="builtin"),
    590: config.get("spectrometer_reading.config", "led_590", fallback="builtin"),
    630: config.get("spectrometer_reading.config", "led_630", fallback="builtin"),
    680: config.get("spectrometer_reading.config", "led_680", fallback="builtin"),
}

def parser(topic: str, payload: pt.MQTTMessagePayload) -> dict:
    metadata = produce_metadata(topic)

    return {
        "experiment": metadata.experiment,
        "pioreactor_unit": metadata.pioreactor_unit,
        "timestamp": current_utc_datetime(),
        "reading": float(payload),
        "band": int(metadata.rest_of_topic[-1].removeprefix("band_")),
    }

register_source_to_sink(
    [
        TopicToParserToTable(
            f"pioreactor/+/+/spectrometer_reading/band_{band}",
            parser,
            "as7341_spectrum_readings",
        ) for band in WAVELENGTH_LED_MAP.keys()
    ]
)

class SpectrometerReading(BackgroundJobWithDodgingContrib):

    job_name = "spectrometer_reading"

    published_settings = {
        f"band_{band}": {"datatype": "float", "unit": "AU", "settable": False}
        for band in WAVELENGTH_LED_MAP.keys()
    }

    def __init__(self, unit: str, experiment: str) -> None:
        super().__init__(unit=unit, experiment=experiment, plugin_name="spectrometer_reading_plugin")

        try:
            i2c = board.I2C()
            self.sensor = adafruit_as7341.AS7341(i2c)
        except Exception:
            self.logger.error("Is the AS7341 board attached to the Pioreactor HAT?")
            self.clean_up()
            raise HardwareNotFoundError("Is the AS7341 board attached to the Pioreactor HAT?")

        self.sensor.led_current = config.getfloat("spectrometer_reading.config", "led_current_mA")
        self.sensor.gain = 10  # use max gain - vary the LED current to avoid saturation
        self.is_setup = False
        self._background_noise = [0.0] * 8

    def record_all_bands(self) -> list[float]:
        normalized_channels = []
        
        for wavelength, channel in WAVELENGTH_LED_MAP.items():
            if channel != "builtin":
                # Turn on the specific LED for this wavelength
                led_utils.led_intensity(
                    {channel: config.getfloat("spectrometer_reading.config", f"led_{wavelength}_intensity", fallback=100)},
                    unit=self.unit,
                    experiment=self.experiment,
                    source_of_event=self.job_name,
                    pubsub_client=self.pub_client,
                )
            else:
                # Use the built-in LED
                self.sensor.led = True

            # Take the reading
            raw_channel = self.sensor.channel_415 if wavelength == 415 else \
                          self.sensor.channel_445 if wavelength == 445 else \
                          self.sensor.channel_480 if wavelength == 480 else \
                          self.sensor.channel_515 if wavelength == 515 else \
                          self.sensor.channel_555 if wavelength == 555 else \
                          self.sensor.channel_590 if wavelength == 590 else \
                          self.sensor.channel_630 if wavelength == 630 else \
                          self.sensor.channel_680
            
            normalized_channel = self.normalize_by_gain_time([raw_channel])[0]
            normalized_channel = self.normalize_by_offset(normalized_channel, wavelength // 5 - 83)
            normalized_channels.append(normalized_channel)

            # Turn off the LED
            if channel != "builtin":
                led_utils.led_intensity(
                    {channel: 0.0},
                    unit=self.unit,
                    experiment=self.experiment,
                    source_of_event=self.job_name,
                    pubsub_client=self.pub_client,
                )
            else:
                self.sensor.led = False

            # Set the attribute for this band
            setattr(self, f"band_{wavelength}", normalized_channel)

        return normalized_channels

    def normalize_by_offset(self, band_recording: float, index: int) -> float:
        return band_recording - self._background_noise[index]

    def normalize_by_gain_time(self, band_recordings: list[int]) -> list[float]:
        return [x / 2 ** (self.sensor.gain - 1) / self.sensor.atime for x in band_recordings]

    def action_to_do_before_od_reading(self):
        # Turn off all LEDs before OD reading
        led_utils.led_intensity(
            {channel: 0.0 for channel in led_utils.ALL_LED_CHANNELS},
            unit=self.unit,
            experiment=self.experiment,
            source_of_event=self.job_name,
            pubsub_client=self.pub_client,
        )
        self.sensor.led = False

    def on_disconnected(self) -> None:
        self.turn_off_led()

    def turn_on_led(self) -> None:
        # This method is not used anymore, as we control individual LEDs in record_all_bands
        pass

    def turn_off_led(self) -> None:
        # Turn off all LEDs
        led_utils.led_intensity(
            {channel: 0.0 for channel in led_utils.ALL_LED_CHANNELS},
            unit=self.unit,
            experiment=self.experiment,
            source_of_event=self.job_name,
            pubsub_client=self.pub_client,
        )
        self.sensor.led = False

    def record_background_noise(self) -> None:
        self.turn_off_led()
        # Initially we record all sensors with LED off, to account for dark current, ambient light, etc.
        self._background_noise = self.normalize_by_gain_time(list(self.sensor.all_channels))
        self.logger.debug(f"Setup done, {self._background_noise=}")

    def action_to_do_after_od_reading(self) -> None:
        with led_utils.change_leds_intensities_temporarily(
            {ch: 0.0 for ch in led_utils.ALL_LED_CHANNELS},
            unit=self.unit,
            experiment=self.experiment,
            source_of_event=self.job_name,
            pubsub_client=self.pub_client,
            verbose=False,
        ):
            if self.is_setup is False:
                self.record_background_noise()
                self.is_setup = True
            else:
                self.record_all_bands()

@click.command(name="spectrometer_reading")
def click_spectrometer_reading() -> None:
    """
    Start spectrometer reading from the AS7341 sensor.
    """
    job = SpectrometerReading(
        unit=get_unit_name(),
        experiment=get_latest_experiment_name(),
    )
    job.block_until_disconnected()
